@@ -32,6 +32,10 @@ enum stk_envelope_type {
 	STK_ENVELOPE_TYPE_EVENT_DOWNLOAD =	0xD6,
 	STK_ENVELOPE_TYPE_TIMER_EXPIRATION =	0xD7,
 	STK_ENVELOPE_TYPE_USSD_DOWNLOAD =	0xD9,
+	STK_ENVELOPE_TYPE_MMS_TRANSFER_STATUS =	0xDA,
+	STK_ENVELOPE_TYPE_MMS_NOTIFICATION =	0xDB,
+	STK_ENVELOPE_TYPE_TERMINAL_APP =	0xDC,
+	STK_ENVELOPE_TYPE_GEOLOCATION_REPORT =	0xDD,
 };
 
 /* TS 102.223 Section 9.4 */
@@ -152,8 +156,10 @@ enum stk_data_object_type {
 	STK_DATA_OBJECT_TYPE_SERVICE_AVAILABILITY =		0x45,
 	STK_DATA_OBJECT_TYPE_ESN =				0x46,
 	STK_DATA_OBJECT_TYPE_NETWORK_ACCESS_NAME =		0x47,
-	STK_DATA_OBJECT_TYPE_CDMA_SMS_TPDU = 			0x48,
+	STK_DATA_OBJECT_TYPE_CDMA_SMS_TPDU =			0x48,
 	STK_DATA_OBJECT_TYPE_REMOTE_ENTITY_ADDRESS =		0x49,
+	STK_DATA_OBJECT_TYPE_I_WLAN_ID_TAG =			0x4A,
+	STK_DATA_OBJECT_TYPE_I_WLAN_ACCESS_STATUS =		0x4B,
 	STK_DATA_OBJECT_TYPE_TEXT_ATTRIBUTE =			0x50,
 	STK_DATA_OBJECT_TYPE_ITEM_TEXT_ATTRIBUTE_LIST =		0x51,
 	STK_DATA_OBJECT_TYPE_PDP_ACTIVATION_PARAMETER =		0x52,
@@ -164,6 +170,7 @@ enum stk_data_object_type {
 	STK_DATA_OBJECT_TYPE_FRAME_LAYOUT =			0x66,
 	STK_DATA_OBJECT_TYPE_FRAMES_INFO =			0x67,
 	STK_DATA_OBJECT_TYPE_FRAME_ID =				0x68,
+	STK_DATA_OBJECT_TYPE_UTRAN_MEASUREMENT_QUALIFIER =	0x69,
 	STK_DATA_OBJECT_TYPE_MMS_REFERENCE =			0x6A,
 	STK_DATA_OBJECT_TYPE_MMS_ID =				0x6B,
 	STK_DATA_OBJECT_TYPE_MMS_TRANSFER_STATUS =		0x6C,
@@ -174,34 +181,37 @@ enum stk_data_object_type {
 	STK_DATA_OBJECT_TYPE_REGISTRY_APPLICATION_DATA =	0x71,
 	STK_DATA_OBJECT_TYPE_ROUTING_AREA_INFO =		0x73,
 	STK_DATA_OBJECT_TYPE_UPDATE_ATTACH_TYPE =		0x74,
+	STK_DATA_OBJECT_TYPE_REJECTION_CAUSE_CODE =		0x75,
 	STK_DATA_OBJECT_TYPE_NMEA_SENTENCE =			0x78,
 	STK_DATA_OBJECT_TYPE_PLMN_LIST =			0x79,
 	STK_DATA_OBJECT_TYPE_BROADCAST_NETWORK_INFO =		0x7A,
 	STK_DATA_OBJECT_TYPE_ACTIVATE_DESCRIPTOR =		0x7B,
+	STK_DATA_OBJECT_TYPE_EPS_PDN_CONN_ACTIVATION_REQ =	0x7C,
+	STK_DATA_OBJECT_TYPE_TRACKING_AREA_ID =			0x7D,
 };
 
 enum stk_device_identity_type {
-	STK_DEVICE_IDENTITY_TYPE_KEYPAD =	0x01,
-	STK_DEVICE_IDENTITY_TYPE_DISPLAY =	0x02,
-	STK_DEVICE_IDENTITY_TYPE_EARPIECE =	0x03,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_0 = 0x10,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_1 = 0x11,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_2 = 0x12,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_3 = 0x13,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_4 = 0x14,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_5 = 0x15,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_6 = 0x16,
-	STK_DEVICE_IDENTITY_TYPE_CARD_READER_7 = 0x17,
-	STK_DEVICE_IDENTITY_TYPE_CHANNEL_1 = 0x21,
-	STK_DEVICE_IDENTITY_TYPE_CHANNEL_7 = 0x27,
-	STK_DEVICE_IDENTITY_TYPE_UICC =		0x81,
-	STK_DEVICE_IDENTITY_TYPE_TERMINAL =	0x82,
-	STK_DEVICE_IDENTITY_TYPE_NETWORK =	0x83,
+	STK_DEVICE_IDENTITY_TYPE_KEYPAD =		0x01,
+	STK_DEVICE_IDENTITY_TYPE_DISPLAY =		0x02,
+	STK_DEVICE_IDENTITY_TYPE_EARPIECE =		0x03,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_0 =	0x10,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_1 =	0x11,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_2 =	0x12,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_3 =	0x13,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_4 =	0x14,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_5 =	0x15,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_6 =	0x16,
+	STK_DEVICE_IDENTITY_TYPE_CARD_READER_7 =	0x17,
+	STK_DEVICE_IDENTITY_TYPE_CHANNEL_1 =		0x21,
+	STK_DEVICE_IDENTITY_TYPE_CHANNEL_7 =		0x27,
+	STK_DEVICE_IDENTITY_TYPE_UICC =			0x81,
+	STK_DEVICE_IDENTITY_TYPE_TERMINAL =		0x82,
+	STK_DEVICE_IDENTITY_TYPE_NETWORK =		0x83,
 };
 
 enum stk_qualifier_get_reader_status_type {
-	STK_QUALIFIER_TYPE_CARD_READER_STATUS = 	0x00,
-	STK_QUALIFIER_TYPE_CARD_READER_ID = 		0x01,
+	STK_QUALIFIER_TYPE_CARD_READER_STATUS =		0x00,
+	STK_QUALIFIER_TYPE_CARD_READER_ID =		0x01,
 };
 
 enum stk_duration_type {
@@ -307,68 +317,68 @@ enum stk_event_type {
 };
 
 enum stk_service_state {
-	STK_NORMAL_SERVICE = 	0x00,
-	STK_LIMITED_SERVICE = 	0x01,
-	STK_NO_SERVICE = 	0x02
+	STK_NORMAL_SERVICE =	0x00,
+	STK_LIMITED_SERVICE =	0x01,
+	STK_NO_SERVICE =	0x02
 };
 
 enum stk_icon_qualifier {
-	STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY = 	0x00,
+	STK_ICON_QUALIFIER_TYPE_SELF_EXPLANATORY =	0x00,
 	STK_ICON_QUALIFIER_TYPE_NON_SELF_EXPLANATORY =	0x01
 };
 
 enum stk_ins {
-	STK_INS_DEACTIVATE_FILE = 			0x04,
-	STK_INS_ERASE_RECORDS = 			0x0C,
-	STK_INS_ERASE_BINARY_0E = 			0x0E,
-	STK_INS_ERASE_BINARY_0F = 			0x0F,
-	STK_INS_PERFORM_SCQL_OPERATION = 		0x10,
-	STK_INS_PERFORM_TRANSACTION_OPERATION = 	0x12,
-	STK_INS_PERFORM_USER_OPERATION = 		0x14,
-	STK_INS_VERIFY_20 = 				0x20,
-	STK_INS_VERIFY_21 = 				0x21,
-	STK_INS_MANAGE_SECURITY_ENVIRONMENT = 		0x22,
-	STK_INS_CHANGE_REFERENCE_DATA = 		0x24,
-	STK_INS_DISABLE_VERIFICATION_REQUIREMENT = 	0x26,
-	STK_INS_ENABLE_VERIFICATION_REQUIREMENT = 	0x28,
-	STK_INS_PERFORM_SECURITY_OPERATION = 		0x2A,
-	STK_INS_RESET_RETRY_COUNTER = 			0x2C,
-	STK_INS_ACTIVATE_FILE = 			0x44,
-	STK_INS_GENERATE_ASYMMETRIC_KEY_PAIR = 		0x46,
-	STK_INS_MANAGE_CHANNEL = 			0x70,
-	STK_INS_EXTERNAL_AUTHENTICATE = 		0x82,
-	STK_INS_GET_CHALLENGE = 			0x84,
-	STK_INS_GENERAL_AUTHENTICATE_86 = 		0x86,
-	STK_INS_GENERAL_AUTHENTICATE_87 = 		0x87,
-	STK_INS_INTERNAL_AUTHENTICATE = 		0x88,
-	STK_INS_SEARCH_BINARY_A0 = 			0xA0,
-	STK_INS_SEARCH_BINARY_A1 = 			0xA1,
-	STK_INS_SEARCH_RECORD = 			0xA2,
-	STK_INS_SELECT = 				0xA4,
-	STK_INS_READ_BINARY_B0 = 			0xB0,
-	STK_INS_READ_BINARY_B1 = 			0xB1,
-	STK_INS_READ_RECORDS_B2 = 			0xB2,
-	STK_INS_READ_RECORDS_B3 = 			0xB3,
-	STK_INS_GET_RESPONSE = 				0xC0,
-	STK_INS_ENVELOPE_C2 = 				0xC2,
+	STK_INS_DEACTIVATE_FILE =			0x04,
+	STK_INS_ERASE_RECORDS =				0x0C,
+	STK_INS_ERASE_BINARY_0E =			0x0E,
+	STK_INS_ERASE_BINARY_0F =			0x0F,
+	STK_INS_PERFORM_SCQL_OPERATION =		0x10,
+	STK_INS_PERFORM_TRANSACTION_OPERATION =		0x12,
+	STK_INS_PERFORM_USER_OPERATION =		0x14,
+	STK_INS_VERIFY_20 =				0x20,
+	STK_INS_VERIFY_21 =				0x21,
+	STK_INS_MANAGE_SECURITY_ENVIRONMENT =		0x22,
+	STK_INS_CHANGE_REFERENCE_DATA =			0x24,
+	STK_INS_DISABLE_VERIFICATION_REQUIREMENT =	0x26,
+	STK_INS_ENABLE_VERIFICATION_REQUIREMENT =	0x28,
+	STK_INS_PERFORM_SECURITY_OPERATION =		0x2A,
+	STK_INS_RESET_RETRY_COUNTER =			0x2C,
+	STK_INS_ACTIVATE_FILE =				0x44,
+	STK_INS_GENERATE_ASYMMETRIC_KEY_PAIR =		0x46,
+	STK_INS_MANAGE_CHANNEL =			0x70,
+	STK_INS_EXTERNAL_AUTHENTICATE =			0x82,
+	STK_INS_GET_CHALLENGE =				0x84,
+	STK_INS_GENERAL_AUTHENTICATE_86 =		0x86,
+	STK_INS_GENERAL_AUTHENTICATE_87 =		0x87,
+	STK_INS_INTERNAL_AUTHENTICATE =			0x88,
+	STK_INS_SEARCH_BINARY_A0 =			0xA0,
+	STK_INS_SEARCH_BINARY_A1 =			0xA1,
+	STK_INS_SEARCH_RECORD =				0xA2,
+	STK_INS_SELECT =				0xA4,
+	STK_INS_READ_BINARY_B0 =			0xB0,
+	STK_INS_READ_BINARY_B1 =			0xB1,
+	STK_INS_READ_RECORDS_B2 =			0xB2,
+	STK_INS_READ_RECORDS_B3 =			0xB3,
+	STK_INS_GET_RESPONSE =				0xC0,
+	STK_INS_ENVELOPE_C2 =				0xC2,
 	STK_INS_ENVELOPE_C3 =				0xC3,
 	STK_INS_GET_DATA_CA =				0xCA,
 	STK_INS_GET_DATA_CB =				0xCB,
 	STK_INS_WRITE_BINARY_D0 =			0xD0,
-	STK_INS_WRITE_BINARY_D1 = 			0xD1,
-	STK_INS_WRITE_RECORD = 				0xD2,
-	STK_INS_UPDATE_BINARY_D6 = 			0xD6,
-	STK_INS_UPDATE_BINARY_D7 = 			0xD7,
-	STK_INS_PUT_DATA_DA = 				0xDA,
-	STK_INS_PUT_DATA_DB = 				0xDB,
-	STK_INS_UPDATE_RECORD_DC = 			0xDC,
-	STK_INS_UPDATE_RECORD_DD = 			0xDD,
-	STK_INS_CREATE_FILE = 				0xE0,
-	STK_INS_APPEND_RECORD = 			0xE2,
-	STK_INS_DELETE_FILE = 				0xE4,
-	STK_INS_TERMINATE_DF = 				0xE6,
-	STK_INS_TERMINATE_EF = 				0xE8,
-	STK_INS_TERMINATE_CARD_USAGE = 			0xFE
+	STK_INS_WRITE_BINARY_D1 =			0xD1,
+	STK_INS_WRITE_RECORD =				0xD2,
+	STK_INS_UPDATE_BINARY_D6 =			0xD6,
+	STK_INS_UPDATE_BINARY_D7 =			0xD7,
+	STK_INS_PUT_DATA_DA =				0xDA,
+	STK_INS_PUT_DATA_DB =				0xDB,
+	STK_INS_UPDATE_RECORD_DC =			0xDC,
+	STK_INS_UPDATE_RECORD_DD =			0xDD,
+	STK_INS_CREATE_FILE =				0xE0,
+	STK_INS_APPEND_RECORD =				0xE2,
+	STK_INS_DELETE_FILE =				0xE4,
+	STK_INS_TERMINATE_DF =				0xE6,
+	STK_INS_TERMINATE_EF =				0xE8,
+	STK_INS_TERMINATE_CARD_USAGE =			0xFE
 };
 
 enum stk_browser_id {
@@ -380,64 +390,65 @@ enum stk_browser_id {
 };
 
 enum stk_bearer {
-	STK_BEARER_SMS = 	0x00,
+	STK_BEARER_SMS =	0x00,
 	STK_BEARER_CS_DATA =	0x01,
 	STK_BEARER_GSM_3G =	0x02,
-	STK_BEARER_PS = 	0x03
+	STK_BEARER_PS =		0x03
 };
 
 enum stk_browser_termination_cause {
-	STK_BROWSER_USER_TERMINATION = 		0x00,
-	STK_BROWSER_ERROR_TERMINATION = 	0x01
+	STK_BROWSER_USER_TERMINATION =		0x00,
+	STK_BROWSER_ERROR_TERMINATION =		0x01
 };
 
 enum stk_bearer_type {
-	STK_BEARER_TYPE_DEFAULT = 		0x03,
-	STK_BEARER_TYPE_INDEPENDENT = 		0x04,
+	STK_BEARER_TYPE_DEFAULT =		0x03,
+	STK_BEARER_TYPE_INDEPENDENT =		0x04,
 	STK_BEARER_TYPE_BLUETOOTH =		0x05,
 	STK_BEARER_TYPE_IRDA =			0x06,
-	STK_BEARER_TYPE_RS232 = 		0x07,
-	STK_BEARER_TYPE_PACKET_DATA_SERVICE = 	0x08,
-	STK_BEARER_TYPE_I_WLAN = 		0x0a,
-	STK_BEARER_TYPE_USB = 			0x10
+	STK_BEARER_TYPE_RS232 =			0x07,
+	STK_BEARER_TYPE_PACKET_DATA_SERVICE =	0x08,
+	STK_BEARER_TYPE_I_WLAN =		0x0a,
+	STK_BEARER_TYPE_USB =			0x10
 };
 
 enum stk_address_type {
-	STK_ADDRESS_IPV4 = 	0x21,
-	STK_ADDRESS_IPV6 = 	0x57
+	STK_ADDRESS_AUTO =	-1,
+	STK_ADDRESS_IPV4 =	0x21,
+	STK_ADDRESS_IPV6 =	0x57
 };
 
 enum stk_access_technology_type {
-	STK_ACCESS_TECHNOLOGY_GSM = 		0x00,
+	STK_ACCESS_TECHNOLOGY_GSM =		0x00,
 	STK_ACCESS_TECHNOLOGY_TIA_EIA_553 =	0x01,
 	STK_ACCESS_TECHNOLOGY_TIA_EIA_136_C =	0x02,
-	STK_ACCESS_TECHNOLOGY_UTRAN = 		0x03,
-	STK_ACCESS_TECHNOLOGY_TETRA = 		0x04,
-	STK_ACCESS_TECHNOLOGY_TIA_EIA_95 = 	0x05,
+	STK_ACCESS_TECHNOLOGY_UTRAN =		0x03,
+	STK_ACCESS_TECHNOLOGY_TETRA =		0x04,
+	STK_ACCESS_TECHNOLOGY_TIA_EIA_95 =	0x05,
 	STK_ACCESS_TECHNOLOGY_CDMA2000_1X =	0x06,
-	STK_ACCESS_TECHNOLOGY_CDMA2000_HRPD = 	0x07,
-	STK_ACCESS_TECHNOLOGY_EUTRAN = 		0x08
+	STK_ACCESS_TECHNOLOGY_CDMA2000_HRPD =	0x07,
+	STK_ACCESS_TECHNOLOGY_EUTRAN =		0x08
 };
 
 enum stk_technology_id {
-	STK_TECHNOLOGY_INDEPENDENT = 	0x00,
-	STK_TECHNOLOGY_BLUETOOTH = 	0x01,
-	STK_TECHNOLOGY_IRDA = 		0x02,
-	STK_TECHNOLOGY_RS232 = 		0x03,
-	STK_TECHNOLOGY_USB = 		0x04
+	STK_TECHNOLOGY_INDEPENDENT =	0x00,
+	STK_TECHNOLOGY_BLUETOOTH =	0x01,
+	STK_TECHNOLOGY_IRDA =		0x02,
+	STK_TECHNOLOGY_RS232 =		0x03,
+	STK_TECHNOLOGY_USB =		0x04
 };
 
 enum stk_battery_state {
-	STK_BATTERY_VERY_LOW = 	0x00,
-	STK_BATTERY_LOW = 	0x01,
-	STK_BATTERY_AVERAGE = 	0x02,
-	STK_BATTERY_GOOD = 	0x03,
-	STK_BATTERY_FULL = 	0x04
+	STK_BATTERY_VERY_LOW =	0x00,
+	STK_BATTERY_LOW =	0x01,
+	STK_BATTERY_AVERAGE =	0x02,
+	STK_BATTERY_GOOD =	0x03,
+	STK_BATTERY_FULL =	0x04
 };
 
 enum stk_frame_layout_type {
-	STK_LAYOUT_HORIZONTAL = 	0x01,
-	STK_LAYOUT_VERTICAL = 		0x02
+	STK_LAYOUT_HORIZONTAL =		0x01,
+	STK_LAYOUT_VERTICAL =		0x02
 };
 
 enum stk_broadcast_network_technology {
@@ -447,6 +458,111 @@ enum stk_broadcast_network_technology {
 	STK_BROADCAST_NETWORK_T_DMB = 0x03
 };
 
+enum stk_i_wlan_access_status {
+	STK_I_WLAN_STATUS_NO_COVERAGE		= 0x00,
+	STK_I_WLAN_STATUS_NOT_CONNECTED		= 0x01,
+	STK_I_WLAN_STATUS_CONNECTED		= 0x02,
+};
+
+enum stk_update_attach_type {
+	STK_UPDATE_ATTACH_NORMAL_LOCATION_UPDATING	= 0x00,
+	STK_UPDATE_ATTACH_PERIODIC_UPDATING		= 0x01,
+	STK_UPDATE_ATTACH_IMSI_ATTACH			= 0x02,
+	STK_UPDATE_ATTACH_GPRS_ATTACH			= 0x03,
+	STK_UPDATE_ATTACH_GPRS_IMSI_ATTACH		= 0x04,
+	STK_UPDATE_ATTACH_RA_UPDATING			= 0x05,
+	STK_UPDATE_ATTACH_RA_LA_UPDATING		= 0x06,
+	STK_UPDATE_ATTACH_RA_LA_UPDATING_IMSI_ATTACH	= 0x07,
+	STK_UPDATE_ATTACH_PERIODIC_RA_UPDATING		= 0x08,
+	STK_UPDATE_ATTACH_EPS_ATTACH			= 0x09,
+	STK_UPDATE_ATTACH_EPS_IMSI_ATTACH		= 0x0a,
+	STK_UPDATE_ATTACH_TA_UPDATING			= 0x0b,
+	STK_UPDATE_ATTACH_TA_LA_UPDATING		= 0x0c,
+	STK_UPDATE_ATTACH_TA_LA_UPDATING_IMSI_ATTACH	= 0x0d,
+	STK_UPDATE_ATTACH_PERIDIC_TA_UPDATING		= 0x0e,
+};
+
+enum stk_rejection_cause_code {
+	/* MM and GMM codes (GERAN/UTRAN) */
+	STK_CAUSE_GMM_IMSI_UNKNOWN_IN_HLR		= 0x02,
+	STK_CAUSE_GMM_ILLEGAL_MS			= 0x03,
+	STK_CAUSE_GMM_IMSI_UNKNOWN_IN_VLR		= 0x04,
+	STK_CAUSE_GMM_IMEI_NOT_ACCEPTED			= 0x05,
+	STK_CAUSE_GMM_ILLEGAL_ME			= 0x06,
+	STK_CAUSE_GMM_GPRS_NOT_ALLOWED			= 0x07,
+	STK_CAUSE_GMM_GPRS_AND_NON_GPRS_NOT_ALLOWED	= 0x08,
+	STK_CAUSE_GMM_IMEI_NOT_DERIVED_BY_NETWORK	= 0x09,
+	STK_CAUSE_GMM_IMPLICITLY_DETACHED		= 0x0a,
+	STK_CAUSE_GMM_PLMN_NOT_ALLOWED			= 0x0b,
+	STK_CAUSE_GMM_LAC_NOT_ALLOWED			= 0x0c,
+	STK_CAUSE_GMM_ROAMING_NOT_ALLOWED		= 0x0d,
+	STK_CAUSE_GMM_GPRS_NOT_ALLOWED_IN_PLMN		= 0x0e,
+	STK_CAUSE_GMM_NO_SUITABLE_CELLS			= 0x0f,
+	STK_CAUSE_GMM_MSC_TEMPORARILY_UNREACHABLE	= 0x10,
+	STK_CAUSE_GMM_NETWORK_FAILURE			= 0x11,
+	STK_CAUSE_GMM_MAC_FAILURE			= 0x14,
+	STK_CAUSE_GMM_SYNCH_FAILURE			= 0x15,
+	STK_CAUSE_GMM_CONGESTION			= 0x16,
+	STK_CAUSE_GMM_GSM_AUTHENTICATION_UNACCEPTABLE	= 0x17,
+	STK_CAUSE_GMM_NOT_AUTHORISED_FOR_CSG		= 0x19,
+	STK_CAUSE_GMM_SERVICE_OPTION_NOT_SUPPORTED	= 0x20,
+	STK_CAUSE_GMM_SERVICE_OPTION_NOT_SUBSCRIBED	= 0x21,
+	STK_CAUSE_GMM_SERVICE_OPTION_TEMPORARY_DEFUNC	= 0x22,
+	STK_CAUSE_GMM_CALL_NOT_IDENTIFIED		= 0x26,
+	STK_CAUSE_GMM_NO_PDP_CONTEXT_ACTIVATED		= 0x28,
+	STK_CAUSE_GMM_RETRY_ON_NEW_CELL			= 0x30, /* to 0x3f */
+	STK_CAUSE_GMM_SEMANTICALLY_INCORRECT_MESSAGE	= 0x5f,
+	STK_CAUSE_GMM_INVALID_MANDATORY_INFO		= 0x60,
+	STK_CAUSE_GMM_MESSAGE_TYPE_UNKNOWN		= 0x61,
+	STK_CAUSE_GMM_MESSAGE_TYPE_INCOMPATIBLE_STATE	= 0x62,
+	STK_CAUSE_GMM_IE_UNKNOWN			= 0x63,
+	STK_CAUSE_GMM_CONDITIONAL_IE_ERROR		= 0x64,
+	STK_CAUSE_GMM_MESSAGE_INCOMPATIBLE_WITH_STATE	= 0x65,
+	STK_CAUSE_GMM_PROTOCOL_ERROR			= 0x6f,
+	/* EMM codes (E-UTRAN) */
+	STK_CAUSE_EMM_IMSI_UNKNOWN_IN_HSS		= 0x02,
+	STK_CAUSE_EMM_ILLEGAL_UE			= 0x03,
+	STK_CAUSE_EMM_ILLEGAL_ME			= 0x06,
+	STK_CAUSE_EMM_EPS_NOT_ALLOWED			= 0x07,
+	STK_CAUSE_EMM_EPS_AND_NON_EPS_NOT_ALLOWED	= 0x08,
+	STK_CAUSE_EMM_IMEI_NOT_DERIVED_BY_NETWORK	= 0x09,
+	STK_CAUSE_EMM_IMPLICITLY_DETACHED		= 0x0a,
+	STK_CAUSE_EMM_PLMN_NOT_ALLOWED			= 0x0b,
+	STK_CAUSE_EMM_TAC_NOT_ALLOWED			= 0x0c,
+	STK_CAUSE_EMM_ROAMING_NOT_ALLOWED		= 0x0d,
+	STK_CAUSE_EMM_EPS_NOT_ALLOWED_IN_PLMN		= 0x0e,
+	STK_CAUSE_EMM_NO_SUITABLE_CELLS			= 0x0f,
+	STK_CAUSE_EMM_MSC_TEMPORARILY_UNREACHABLE	= 0x10,
+	STK_CAUSE_EMM_NETWORK_FAILURE			= 0x11,
+	STK_CAUSE_EMM_MAC_FAILURE			= 0x14,
+	STK_CAUSE_EMM_SYNCH_FAILURE			= 0x15,
+	STK_CAUSE_EMM_CONGESTION			= 0x16,
+	STK_CAUSE_EMM_SECURITY_MODE_REJECTED		= 0x18,
+	STK_CAUSE_EMM_NOT_AUTHORISED_FOR_CSG		= 0x19,
+	STK_CAUSE_EMM_CS_FALLBACK_NOT_ALLOWED		= 0x26,
+	STK_CAUSE_EMM_CS_DOMAIN_TEMPORARY_UNAVAILABLE	= 0x27,
+	STK_CAUSE_EMM_NO_EPS_BEARER_CONTEXT_ACTIVATED	= 0x28,
+	STK_CAUSE_EMM_SEMANTICALLY_INCORRECT_MESSAGE	= 0x5f,
+	STK_CAUSE_EMM_INVALID_MANDATORY_INFO		= 0x60,
+	STK_CAUSE_EMM_MESSAGE_TYPE_UNKNOWN		= 0x61,
+	STK_CAUSE_EMM_MESSAGE_TYPE_INCOMPATIBLE_STATE	= 0x62,
+	STK_CAUSE_EMM_IE_UNKNOWN			= 0x63,
+	STK_CAUSE_EMM_CONDITIONAL_IE_ERROR		= 0x64,
+	STK_CAUSE_EMM_MESSAGE_INCOMPATIBLE_WITH_STATE	= 0x65,
+	STK_CAUSE_EMM_PROTOCOL_ERROR			= 0x6f,
+};
+
+enum stk_me_status {
+	STK_ME_STATUS_IDLE = 		0x00,
+	STK_ME_STATUS_NOT_IDLE = 	0x01
+};
+
+enum stk_img_scheme {
+	STK_IMG_SCHEME_BASIC =		0x11,
+	STK_IMG_SCHEME_COLOR =		0x21,
+	STK_IMG_SCHEME_TRANSPARENCY =	0x22,
+};
+
 /* For data object that only has a byte array with undetermined length */
 struct stk_common_byte_array {
 	unsigned char *array;
@@ -474,6 +590,7 @@ struct stk_address {
  * bytes."
  */
 struct stk_subaddress {
+	ofono_bool_t has_subaddr;
 	unsigned char len;
 	unsigned char subaddr[23];
 };
@@ -494,6 +611,12 @@ struct stk_ccp {
 	unsigned char ccp[16];
 };
 
+/* Defined in TS 31.111 Section 8.5 */
+struct stk_cbs_page {
+	unsigned char len;
+	unsigned char page[88];
+};
+
 /*
  * According to 102.223 Section 8.8 interval values of 0x00 are reserved.
  * We use this to denote empty duration objects.
@@ -525,6 +648,19 @@ struct stk_result {
 	unsigned char *additional;
 };
 
+/* Defined in TS 102.223 Section 8.14 */
+struct stk_ss {
+	unsigned char ton_npi;
+	char *ss;
+};
+
+/* Defined in TS 131.111 Section 8.17.  Length limit of 160 chars in 23.028 */
+struct stk_ussd_string {
+	unsigned char dcs;
+	unsigned char string[160];
+	int len;
+};
+
 /* Define the struct of single file in TS102.223 Section 8.18.
  * According to TS 11.11 Section 6.2, each file id has two bytes, and the
  * maximum Dedicated File level is 2. So the maximum size of file is 8, which
@@ -585,6 +721,17 @@ struct stk_transaction_id {
 	unsigned int len;
 };
 
+/*
+ * According to 31.111 Section 8.29 the length of CTLV is 1 byte. This means
+ * that the maximum size is 127 according to the rules of CTLVs. Each channel
+ * is represented as 10 bits, so the maximum number of channel is 127*8/10=101.
+ */
+struct stk_bcch_channel_list {
+	unsigned short channels[101];
+	unsigned int num;
+	ofono_bool_t has_list;
+};
+
 /*
  * Defined in TS 102.223 Section 8.31
  * Icon ID denotes a file on the SIM filesystem.  Since EF cannot have record
@@ -607,6 +754,16 @@ struct stk_item_icon_id_list {
 	unsigned int len;
 };
 
+/* Defined in TS 102.223 Section 8.33 */
+struct stk_reader_status {
+	int id;
+	ofono_bool_t removable;
+	ofono_bool_t present;
+	ofono_bool_t id1_size;
+	ofono_bool_t card_present;
+	ofono_bool_t card_powered;
+};
+
 /*
  * According to 102.223 Section 8.34 the length of CTLV is 1 byte. This means
  * that the maximum size is 127 according to the rules of CTLVs.
@@ -649,6 +806,24 @@ struct stk_timer_value {
 	unsigned char second;
 };
 
+/* Defined in TS 102.223 Section 8.42 */
+struct stk_bc_repeat {
+	ofono_bool_t has_bc_repeat;
+	unsigned char value;
+};
+
+/* Defined in TS 31.111 Section 8.46 */
+struct stk_timing_advance {
+	ofono_bool_t has_value;
+	enum stk_me_status status;
+	/*
+	 * Contains bit periods number according to 3GPP TS
+	 * 44.118 Section 9.3.106 / 3GPP TS 44.018 Section
+	 * 10.5.2.40.1, not microseconds
+	 */
+	unsigned char advance;
+};
+
 /*
  * According to 102.223 Section 8.52 the length of CTLV is 1 byte. This means
  * that the maximum size is 127 according to the rules of CTLVs. This size also
@@ -677,10 +852,10 @@ struct stk_card_reader_id {
 struct stk_other_address {
 	union {
 		/* Network Byte Order */
-		unsigned int ipv4;
+		uint32_t ipv4;
 		unsigned char ipv6[16];
 	} addr;
-	unsigned char type;
+	enum stk_address_type type;
 };
 
 /* Defined in TS 102.223 Section 8.59 */
@@ -742,6 +917,7 @@ struct stk_attribute_info {
  */
 struct stk_remote_entity_address {
 	unsigned char coding_type;
+	ofono_bool_t has_address;
 	union {
 		unsigned char ieee802[6];
 		unsigned char irda[4];
@@ -767,6 +943,12 @@ struct stk_text_attribute {
 	unsigned char len;
 };
 
+/* Defined in TS 31.111 Section 8.72 */
+struct stk_pdp_act_par {
+	unsigned char par[127];
+	unsigned char len;
+};
+
 /*
  * According to 102.223 Section 8.73 the length of CTLV is 1 byte. This means
  * that the maximum size is 127 according to the rules of CTLVs. In addition,
@@ -797,7 +979,9 @@ struct stk_frame_layout {
  */
 struct stk_frames_info {
 	unsigned char id;
-	unsigned char list[126];
+	struct {
+		unsigned char width, height;
+	} list[63];
 	unsigned int len;
 };
 
@@ -862,6 +1046,21 @@ struct stk_broadcast_network_information {
 	unsigned int len;
 };
 
+/* Defined in TS 131.111 Section 8.91 */
+struct stk_routing_area_info {
+	char mnc[OFONO_MAX_MNC_LENGTH + 1];
+	char mcc[OFONO_MAX_MCC_LENGTH + 1];
+	unsigned short lac;
+	unsigned char rac;
+};
+
+/* Defined in TS 131.111 Section 8.99 */
+struct stk_tracking_area_id {
+	char mnc[OFONO_MAX_MNC_LENGTH + 1];
+	char mcc[OFONO_MAX_MCC_LENGTH + 1];
+	unsigned short tac;
+};
+
 struct stk_command_display_text {
 	char *text;
 	struct stk_icon_id icon_id;
@@ -925,7 +1124,6 @@ struct stk_command_select_item {
 
 struct stk_command_send_sms {
 	char *alpha_id;
-	struct stk_address address;
 	struct sms gsm_sms;
 	struct stk_common_byte_array cdma_sms;
 	struct stk_icon_id icon_id;
@@ -933,6 +1131,22 @@ struct stk_command_send_sms {
 	struct stk_frame_id frame_id;
 };
 
+struct stk_command_send_ss {
+	char *alpha_id;
+	struct stk_ss ss;
+	struct stk_icon_id icon_id;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_send_ussd {
+	char *alpha_id;
+	struct stk_ussd_string ussd_string;
+	struct stk_icon_id icon_id;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
 struct stk_command_setup_call {
 	char *alpha_id_usr_cfm;
 	struct stk_address addr;
@@ -1011,12 +1225,102 @@ struct stk_command_launch_browser {
 	char *text_passwd;
 };
 
+struct stk_command_close_channel {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_receive_data {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	unsigned char data_len;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_send_data {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	struct stk_common_byte_array data;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_service_search {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	struct stk_service_search serv_search;
+	struct stk_device_filter dev_filter;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_get_service_info {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	struct stk_attribute_info attr_info;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_declare_service {
+	struct stk_service_record serv_rec;
+	struct stk_uicc_te_interface intf;
+};
+
+struct stk_command_set_frames {
+	struct stk_frame_id frame_id;
+	struct stk_frame_layout frame_layout;
+	struct stk_frame_id frame_id_default;
+};
+
+struct stk_command_retrieve_mms {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	struct stk_mms_reference mms_ref;
+	GSList *mms_rec_files;
+	struct stk_mms_content_id mms_content_id;
+	struct stk_mms_id mms_id;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_submit_mms {
+	char *alpha_id;
+	struct stk_icon_id icon_id;
+	GSList *mms_subm_files;
+	struct stk_mms_id mms_id;
+	struct stk_text_attribute text_attr;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_display_mms {
+	GSList *mms_subm_files;
+	struct stk_mms_id mms_id;
+	ofono_bool_t imd_resp;
+	struct stk_frame_id frame_id;
+};
+
+struct stk_command_activate {
+	unsigned char actv_desc;
+};
+
+enum stk_command_parse_result {
+	STK_PARSE_RESULT_OK,
+	STK_PARSE_RESULT_TYPE_NOT_UNDERSTOOD,
+	STK_PARSE_RESULT_DATA_NOT_UNDERSTOOD,
+	STK_PARSE_RESULT_MISSING_VALUE,
+};
+
 struct stk_command {
 	unsigned char number;
 	unsigned char type;
 	unsigned char qualifier;
 	enum stk_device_identity_type src;
 	enum stk_device_identity_type dst;
+	enum stk_command_parse_result status;
 
 	union {
 		struct stk_command_display_text display_text;
@@ -1028,6 +1332,8 @@ struct stk_command {
 		struct stk_command_setup_menu setup_menu;
 		struct stk_command_select_item select_item;
 		struct stk_command_send_sms send_sms;
+		struct stk_command_send_ss send_ss;
+		struct stk_command_send_ussd send_ussd;
 		struct stk_command_setup_call setup_call;
 		struct stk_command_setup_event_list setup_event_list;
 		struct stk_command_perform_card_apdu perform_card_apdu;
@@ -1037,6 +1343,17 @@ struct stk_command {
 		struct stk_command_send_dtmf send_dtmf;
 		struct stk_command_language_notification language_notification;
 		struct stk_command_launch_browser launch_browser;
+		struct stk_command_close_channel close_channel;
+		struct stk_command_receive_data receive_data;
+		struct stk_command_send_data send_data;
+		struct stk_command_service_search service_search;
+		struct stk_command_get_service_info get_service_info;
+		struct stk_command_declare_service declare_service;
+		struct stk_command_set_frames set_frames;
+		struct stk_command_retrieve_mms retrieve_mms;
+		struct stk_command_submit_mms submit_mms;
+		struct stk_command_display_mms display_mms;
+		struct stk_command_activate activate;
 	};
 
 	void (*destructor)(struct stk_command *command);
@@ -1050,7 +1367,8 @@ struct stk_answer_text {
 	char *text;
 	ofono_bool_t packed;
 	ofono_bool_t yesno;
-	/* If a "Yes/No" answer was requested in a GET INKEY command,
+	/*
+	 * If a "Yes/No" answer was requested in a GET INKEY command,
 	 * .yesno must be TRUE and text should be non-NULL to indicate
 	 * a Yes response or NULL to indicate a No response.
 	 */
@@ -1087,25 +1405,13 @@ struct stk_response_local_info {
 		const char *imei;
 		struct stk_network_measurement_results {
 			struct stk_common_byte_array nmr;
-			struct stk_bcch_ch_list {
-				const short *channels;
-				int length;
-			} bcch_ch_list;
+			struct stk_bcch_channel_list bcch_ch_list;
 		} nmr;
 		struct sms_scts datetime;
 		const char *language;
 		enum stk_battery_state battery_charge;
 		enum stk_access_technology_type access_technology;
-		struct stk_timing_advance {
-			enum {
-				STK_TIMING_ADVANCE_ME_IDLE = 0x00,
-				STK_TIMING_ADVANCE_ME_NOT_IDLE = 0x01,
-			} status;
-			/* Contains bit periods number according to 3GPP TS
-			 * 44.118 Section 9.3.106 / 3GPP TS 44.018 Section
-			 * 10.5.2.40.1, not microseconds */
-			int advance;
-		} tadv;
+		struct stk_timing_advance tadv;
 		/* Bits[31:24]: manufacturer, bits[23:0]: serial number */
 		guint32 esn;
 		const char *imeisv;
@@ -1151,6 +1457,7 @@ struct stk_response {
 		struct stk_response_generic display_text;
 		struct stk_response_get_inkey get_inkey;
 		struct stk_response_get_input get_input;
+		struct stk_response_generic more_time;
 		struct stk_response_generic play_tone;
 		struct stk_response_poll_interval poll_interval;
 		struct stk_response_generic refresh;
@@ -1193,6 +1500,127 @@ struct stk_envelope_sms_mo_control {
 	struct stk_location_info location;
 };
 
+enum stk_call_control_type {
+	STK_CC_TYPE_CALL_SETUP,
+	STK_CC_TYPE_SUPPLEMENTARY_SERVICE,
+	STK_CC_TYPE_USSD_OP,
+	STK_CC_TYPE_PDP_CTX_ACTIVATION,
+	STK_CC_TYPE_EPS_PDN_CONNECTION_ACTIVATION,
+};
+
+/* Used both in the ENVELOPE message to UICC and response from UICC */
+struct stk_envelope_call_control {
+	enum stk_call_control_type type;
+	union {
+		struct stk_address address;
+		struct stk_address ss_string;
+		struct stk_ussd_string ussd_string;
+		struct stk_common_byte_array pdp_ctx_params;
+		struct stk_common_byte_array eps_pdn_params;
+	};
+	/*
+	 * At least one of the following two fields must be present in a
+	 * response indicating modification of the call.
+	 * In an EVELOPE message, only allowed for a call setup.
+	 */
+	struct stk_ccp ccp1;
+	struct stk_subaddress subaddress;
+	struct stk_location_info location;
+	/* Only allowed when ccp1 is present */
+	struct stk_ccp ccp2;
+	char *alpha_id;
+	/* Only allowed when both ccp1 and ccp2 are present */
+	struct stk_bc_repeat bc_repeat;
+};
+
+struct stk_envelope_event_download {
+	enum stk_event_type type;
+	union {
+		struct {
+			unsigned char transaction_id;
+			struct stk_address caller_address;
+			struct stk_subaddress caller_subaddress;
+		} mt_call;
+		struct {
+			unsigned char transaction_id;
+		} call_connected;
+		struct {
+			struct stk_transaction_id transaction_ids;
+			struct stk_cause cause;
+		} call_disconnected;
+		struct {
+			enum stk_service_state state;
+			/* Present when state indicated Normal Service */
+			struct stk_location_info info;
+		} location_status;
+		struct stk_reader_status card_reader_status;
+		char language_selection[3];
+		struct {
+			enum stk_browser_termination_cause cause;
+		} browser_termination;
+		struct {
+			unsigned char channel_status[2];
+			unsigned int channel_data_len;
+		} data_available;
+		struct {
+			unsigned char status[2];
+			struct stk_bearer_description bearer_desc;
+			struct stk_other_address address;
+		} channel_status;
+		struct stk_access_technologies access_technology_change;
+		struct stk_display_parameters display_params_changed;
+		struct {
+			/*
+			 * Note the service record subfield is not required,
+			 * only the Technology id and Service id.
+			 */
+			struct stk_service_record service_record;
+			struct stk_remote_entity_address remote_addr;
+			struct stk_uicc_te_interface transport_level;
+			/* Only present if transport_level present */
+			struct stk_other_address transport_addr;
+		} local_connection;
+		enum stk_network_search_mode network_search_mode_change;
+		struct stk_common_byte_array browsing_status;
+		struct stk_frames_info frames_information_change;
+		enum stk_i_wlan_access_status i_wlan_access_status;
+		struct {
+			struct stk_location_info location;
+			struct stk_routing_area_info rai;
+			struct stk_tracking_area_id tai;
+			enum stk_access_technology_type access_tech;
+			enum stk_update_attach_type update_attach;
+			enum stk_rejection_cause_code cause;
+		} network_rejection;
+	};
+};
+
+struct stk_envelope_timer_expiration {
+	unsigned char id;
+	struct stk_timer_value value;
+};
+
+struct stk_envelope_ussd_data_download {
+	struct stk_ussd_string string;
+};
+
+struct stk_envelope_mms_transfer_status {
+	struct stk_file transfer_file;
+	struct stk_mms_id id;
+	struct stk_mms_transfer_status transfer_status;
+};
+
+struct stk_envelope_mms_notification_download {
+	struct stk_common_byte_array msg;
+	ofono_bool_t last;
+};
+
+struct stk_envelope_terminal_apps {
+	struct stk_registry_application_data *list;
+	int count;
+	ofono_bool_t last;
+};
+
 struct stk_envelope {
 	enum stk_envelope_type type;
 	enum stk_device_identity_type src;
@@ -1201,7 +1629,14 @@ struct stk_envelope {
 		struct stk_envelope_sms_pp_download sms_pp_download;
 		struct stk_envelope_cbs_pp_download cbs_pp_download;
 		struct stk_envelope_menu_selection menu_selection;
+		struct stk_envelope_call_control call_control;
 		struct stk_envelope_sms_mo_control sms_mo_control;
+		struct stk_envelope_event_download event_download;
+		struct stk_envelope_timer_expiration timer_expiration;
+		struct stk_envelope_ussd_data_download ussd_data_download;
+		struct stk_envelope_mms_transfer_status mms_status;
+		struct stk_envelope_mms_notification_download mms_notification;
+		struct stk_envelope_terminal_apps terminal_apps;
 	};
 };
 
@@ -1213,3 +1648,8 @@ const unsigned char *stk_pdu_from_response(const struct stk_response *response,
 						unsigned int *out_length);
 const unsigned char *stk_pdu_from_envelope(const struct stk_envelope *envelope,
 						unsigned int *out_length);
+char *stk_text_to_html(const char *text,
+				const unsigned short *attrs, int num_attrs);
+char *stk_image_to_xpm(const unsigned char *img, unsigned int len,
+			enum stk_img_scheme scheme, const unsigned char *clut,
+			unsigned short clut_len);
