@@ -1,7 +1,7 @@
 /*
  * oFono - GSM Telephony Stack for Linux
  *
- * Copyright (C) 2008-2009 Intel Corporation.  All rights reserved.
+ * Copyright (C) 2008-2010 Intel Corporation.  All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -471,34 +471,152 @@ static void test_submit_encode()
 	g_free(encoded_pdu);
 }
 
-static const char *header_test = "0041000B915121551532F40000631A0A031906200A03"
-	"2104100A032705040A032E05080A043807002B8ACD29A85D9ECFC3E7F21C340EBB41E"
-	"3B79B1E4EBB41697A989D1EB340E2379BCC02B1C3F27399059AB7C36C3628EC2683C6"
-	"6FF65B5E2683E8653C1D";
-static int header_test_len = 100;
-static const char *header_test_expected = "EMS messages can contain italic, bold"
-	", large, small and colored text";
+struct text_format_header {
+	unsigned char len;
+	unsigned char start;
+	unsigned char span;
+	unsigned char format;
+	unsigned char color;
+};
 
-static void test_udh_iter()
+struct ems_udh_test {
+	const char *pdu;
+	unsigned int len;
+	const char *expected;
+	unsigned int udl;
+	unsigned int udhl;
+	unsigned int data_len;
+	struct text_format_header formats[];
+};
+
+static struct ems_udh_test ems_udh_test_1 = {
+	.pdu = "0041000B915121551532F40000631A0A031906200A032104100A03270504"
+		"0A032E05080A043807002B8ACD29A85D9ECFC3E7F21C340EBB41E3B79B1"
+		"E4EBB41697A989D1EB340E2379BCC02B1C3F27399059AB7C36C3628EC26"
+		"83C66FF65B5E2683E8653C1D",
+	.len = 100,
+	.expected = "EMS messages can contain italic, bold, large, small and"
+		" colored text",
+	.formats = {
+		{
+			.len = 3,
+			.start = 0x19,
+			.span = 0x06,
+			.format = 0x20,
+		},
+		{
+			.len = 3,
+			.start = 0x21,
+			.span = 0x04,
+			.format = 0x10,
+		},
+		{
+			.len = 3,
+			.start = 0x27,
+			.span = 0x05,
+			.format = 0x04,
+		},
+		{
+			.len = 3,
+			.start = 0x2E,
+			.span = 0x05,
+			.format = 0x08,
+		},
+		{
+			.len = 4,
+			.start = 0x38,
+			.span = 0x07,
+			.format = 0x00,
+			.color = 0x2B,
+		},
+		{
+			.len = 0,
+		}
+	},
+	.udl = 99,
+	.udhl = 26,
+	.data_len = 87,
+};
+
+static struct ems_udh_test ems_udh_test_2 = {
+	.pdu = "079194712272303351030B915121340195F60000FF80230A030F07230A031"
+		"806130A031E0A430A032E0D830A033D14020A035104F60A0355010600159"
+		"D9E83D2735018442FCFE98A243DCC4E97C92C90F8CD26B3407537B92C67A"
+		"7DD65320B1476934173BA3CBD2ED3D1F277FD8C76299CEF3B280C92A7CF6"
+		"83A28CC4E9FDD6532E8FE96935D",
+	.len = 126,
+	.expected = "This is a test\nItalied, bold, underlined, and "
+		"strikethrough.\nNow a right aligned word.",
+	.formats = {
+		{
+			.len = 3,
+			.start = 0x0f,
+			.span = 0x07,
+			.format = 0x23,
+		},
+		{
+			.len = 3,
+			.start = 0x18,
+			.span = 0x06,
+			.format = 0x13,
+		},
+		{
+			.len = 3,
+			.start = 0x1e,
+			.span = 0x0a,
+			.format = 0x43,
+		},
+		{
+			.len = 3,
+			.start = 0x2e,
+			.span = 0x0d,
+			.format = 0x83,
+		},
+		{
+			.len = 3,
+			.start = 0x3d,
+			.span = 0x14,
+			.format = 0x02,
+		},
+		{
+			.len = 3,
+			.start = 0x51,
+			.span = 0x04,
+			.format = 0xf6,
+		},
+		{
+			.len = 3,
+			.start = 0x55,
+			.span = 0x01,
+			.format = 0x06,
+		},
+	},
+	.udl = 128,
+	.udhl = 35,
+	.data_len = 112,
+};
+
+static void test_ems_udh(gconstpointer data)
 {
+	const struct ems_udh_test *test = data;
 	struct sms sms;
 	unsigned char *decoded_pdu;
 	long pdu_len;
 	gboolean ret;
-	int data_len;
-	int udhl;
+	unsigned int data_len;
+	unsigned int udhl;
 	struct sms_udh_iter iter;
 	int max_chars;
 	unsigned char *unpacked;
 	char *utf8;
+	int i;
 
-	decoded_pdu = decode_hex(header_test, -1, &pdu_len, 0);
+	decoded_pdu = decode_hex(test->pdu, -1, &pdu_len, 0);
 
 	g_assert(decoded_pdu);
-	g_assert(pdu_len == (long)strlen(header_test) / 2);
+	g_assert(pdu_len == (long)strlen(test->pdu) / 2);
 
-	ret = sms_decode(decoded_pdu, pdu_len, TRUE,
-				header_test_len, &sms);
+	ret = sms_decode(decoded_pdu, pdu_len, TRUE, test->len, &sms);
 
 	g_free(decoded_pdu);
 
@@ -531,42 +649,47 @@ static void test_udh_iter()
 	}
 
 	udhl = sms.submit.ud[0];
-	g_assert(sms.submit.udl == 99);
-	g_assert(udhl == 26);
+
+	g_assert(sms.submit.udl == test->udl);
+	g_assert(udhl == test->udhl);
 
 	ret = sms_udh_iter_init(&sms, &iter);
 
 	g_assert(ret);
 
-	g_assert(sms_udh_iter_get_ie_type(&iter) == SMS_IEI_TEXT_FORMAT);
-	g_assert(sms_udh_iter_get_ie_length(&iter) == 3);
-	g_assert(sms_udh_iter_has_next(&iter) == TRUE);
-	g_assert(sms_udh_iter_next(&iter) == TRUE);
+	for (i = 0; test->formats[i].len; i++) {
+		if (g_test_verbose()) {
+			int j;
+			unsigned char data[4];
 
-	g_assert(sms_udh_iter_get_ie_type(&iter) == SMS_IEI_TEXT_FORMAT);
-	g_assert(sms_udh_iter_get_ie_length(&iter) == 3);
-	g_assert(sms_udh_iter_has_next(&iter) == TRUE);
-	g_assert(sms_udh_iter_next(&iter) == TRUE);
+			sms_udh_iter_get_ie_data(&iter, data);
 
-	g_assert(sms_udh_iter_get_ie_type(&iter) == SMS_IEI_TEXT_FORMAT);
-	g_assert(sms_udh_iter_get_ie_length(&iter) == 3);
-	g_assert(sms_udh_iter_has_next(&iter) == TRUE);
-	g_assert(sms_udh_iter_next(&iter) == TRUE);
+			g_print("Header:\n");
+			for (j = 0; j < sms_udh_iter_get_ie_length(&iter); j++)
+				g_print("0x%02x ", data[j]);
 
-	g_assert(sms_udh_iter_get_ie_type(&iter) == SMS_IEI_TEXT_FORMAT);
-	g_assert(sms_udh_iter_get_ie_length(&iter) == 3);
-	g_assert(sms_udh_iter_has_next(&iter) == TRUE);
-	g_assert(sms_udh_iter_next(&iter) == TRUE);
+			g_print("\n");
+		}
 
-	g_assert(sms_udh_iter_get_ie_type(&iter) == SMS_IEI_TEXT_FORMAT);
-	g_assert(sms_udh_iter_get_ie_length(&iter) == 4);
-	g_assert(sms_udh_iter_has_next(&iter) == FALSE);
-	g_assert(sms_udh_iter_next(&iter) == FALSE);
-	g_assert(sms_udh_iter_get_ie_type(&iter) == SMS_IEI_INVALID);
+		g_assert(sms_udh_iter_get_ie_type(&iter) ==
+				SMS_IEI_TEXT_FORMAT);
+		g_assert(sms_udh_iter_get_ie_length(&iter) ==
+				test->formats[i].len);
+
+		if (test->formats[i+1].len) {
+			g_assert(sms_udh_iter_has_next(&iter) == TRUE);
+			g_assert(sms_udh_iter_next(&iter) == TRUE);
+		} else {
+			g_assert(sms_udh_iter_has_next(&iter) == FALSE);
+			g_assert(sms_udh_iter_next(&iter) == FALSE);
+			g_assert(sms_udh_iter_get_ie_type(&iter) ==
+					SMS_IEI_INVALID);
+		}
+	}
 
 	data_len = sms_udl_in_bytes(sms.submit.udl, sms.submit.dcs);
 
-	g_assert(data_len == 87);
+	g_assert(data_len == test->data_len);
 
 	max_chars = (data_len - (udhl + 1)) * 8 / 7;
 
@@ -584,7 +707,7 @@ static void test_udh_iter()
 	if (g_test_verbose())
 		g_print("Decoded user data is: %s\n", utf8);
 
-	g_assert(strcmp(utf8, header_test_expected) == 0);
+	g_assert(strcmp(utf8, test->expected) == 0);
 
 	g_free(utf8);
 }
@@ -685,7 +808,7 @@ static void test_assembly()
 	if (g_test_verbose())
 		g_printf("Text:\n%s\n", utf8);
 
-	l = sms_text_prepare(utf8, ref, TRUE, NULL);
+	l = sms_text_prepare(utf8, ref, TRUE, NULL, FALSE);
 	g_assert(l);
 	g_assert(g_slist_length(l) == 3);
 
@@ -715,7 +838,7 @@ static void test_prepare_7bit()
 	int encoded_tpdu_len;
 	char *encoded_pdu;
 
-	r = sms_text_prepare(test_no_fragmentation_7bit, 0, FALSE, NULL);
+	r = sms_text_prepare(test_no_fragmentation_7bit, 0, FALSE, NULL, FALSE);
 
 	g_assert(r != NULL);
 
@@ -762,16 +885,26 @@ static void test_prepare_7bit()
 	g_slist_free(r);
 }
 
-static const char *pad1 = "Shakespeare divided his time between London and Str"
+struct sms_concat_data {
+	const char *str;
+	unsigned int segments;
+};
+
+
+static struct sms_concat_data shakespeare_test = {
+	.str = "Shakespeare divided his time between London and Str"
 	"atford during his career. In 1596, the year before he bought New Plac"
 	"e as his family home in Stratford, Shakespeare was living in the pari"
-	"sh of St. Helen's, Bishopsgate, north of the River Thames.";
+	"sh of St. Helen's, Bishopsgate, north of the River Thames.",
+	.segments = 2,
+};
 
 /* The string in this test should be padded at the end.  This confuses some
  * decoders which do not use udl properly
  */
-static void test_prepare_concat()
+static void test_prepare_concat(gconstpointer data)
 {
+	const struct sms_concat_data *test = data;
 	GSList *r;
 	GSList *l;
 	char *decoded_str;
@@ -786,12 +919,11 @@ static void test_prepare_concat()
 	guint8 seq;
 
 	if (g_test_verbose())
-		g_print("strlen: %zd\n", strlen(pad1));
-
-	r = sms_text_prepare(pad1, 0, TRUE, NULL);
+		g_print("strlen: %zd\n", strlen(test->str));
 
+	r = sms_text_prepare(test->str, 0, TRUE, NULL, FALSE);
 	g_assert(r);
-	g_assert(g_slist_length(r) == 2);
+	g_assert(g_slist_length(r) == test->segments);
 
 	for (l = r; l; l = l->next) {
 		char *strpdu;
@@ -823,6 +955,7 @@ static void test_prepare_concat()
 			g_print("PDU Len: %ld\n", len);
 
 		ok = sms_decode(pdu, len, TRUE, len - 1, &decoded);
+		g_assert(ok);
 
 		if (g_test_verbose())
 			g_print("Pdu udl: %d\n", (int)decoded.submit.udl);
@@ -841,7 +974,7 @@ static void test_prepare_concat()
 		g_printf("Decoded String: %s\n", decoded_str);
 
 	g_assert(decoded_str);
-	g_assert(strcmp(decoded_str, pad1) == 0);
+	g_assert(strcmp(decoded_str, test->str) == 0);
 	g_free(decoded_str);
 	sms_assembly_free(assembly);
 }
@@ -864,7 +997,7 @@ static void test_limit(gunichar uni, int target_size, gboolean use_16bit)
 
 	utf8[i] = '\0';
 
-	l = sms_text_prepare(utf8, 0, use_16bit, NULL);
+	l = sms_text_prepare(utf8, 0, use_16bit, NULL, FALSE);
 
 	g_assert(l);
 	g_assert(g_slist_length(l) == 255);
@@ -877,7 +1010,7 @@ static void test_limit(gunichar uni, int target_size, gboolean use_16bit)
 	memcpy(utf8 + i, utf8_char, stride);
 	utf8[i+stride] = '\0';
 
-	l = sms_text_prepare(utf8, 0, use_16bit, NULL);
+	l = sms_text_prepare(utf8, 0, use_16bit, NULL, FALSE);
 
 	g_assert(l == NULL);
 	g_free(utf8);
@@ -1158,8 +1291,56 @@ static void test_range_minimizer()
 	}
 }
 
+static void test_sr_assembly()
+{
+	const char *sr_pdu1 = "06040D91945152991136F00160124130340A0160124130"
+				"940A00";
+	const char *sr_pdu2 = "06050D91945152991136F00160124130640A0160124130"
+				"450A00";
+        struct sms sr1;
+	struct sms sr2;
+	unsigned char pdu[176];
+	long pdu_len;
+	struct status_report_assembly *sra;
+	gboolean delivered;
+	unsigned int id;
+	struct sms_address addr;
+
+	/* mr 4 & mr 5 */
+
+        decode_hex_own_buf(sr_pdu1, -1, &pdu_len, 0, pdu);
+	g_assert(sms_decode(pdu, pdu_len, FALSE, 26, &sr1) == TRUE);
+
+	decode_hex_own_buf(sr_pdu2, -1, &pdu_len, 0, pdu);
+	g_assert(sms_decode(pdu, pdu_len, FALSE, 26, &sr2) == TRUE);
+
+	g_print("sr1 address: %s, mr: %d\n",
+			sms_address_to_string(&sr1.status_report.raddr),
+			sr1.status_report.mr);
+
+	g_print("sr2 address: %s, mr: %d\n",
+			sms_address_to_string(&sr2.status_report.raddr),
+			sr2.status_report.mr);
+
+	sms_address_from_string(&addr, "+4915259911630");
+
+	sra = status_report_assembly_new(NULL);
+	status_report_assembly_add_fragment(sra, 42, &addr, 4, time(NULL), 2);
+	status_report_assembly_add_fragment(sra, 42, &addr, 5, time(NULL), 2);
+
+	g_assert(!status_report_assembly_report(sra, &sr1, &id, &delivered));
+	g_assert(status_report_assembly_report(sra, &sr2, &id, &delivered));
+
+	g_assert(id == 42);
+	g_assert(delivered == TRUE);
+	status_report_assembly_free(sra);
+}
+
 int main(int argc, char **argv)
 {
+	char long_string[152*33 + 1];
+	struct sms_concat_data long_string_test;
+
 	g_test_init(&argc, &argv, NULL);
 
 	g_test_add_func("/testsms/Test Simple Deliver", test_simple_deliver);
@@ -1167,10 +1348,30 @@ int main(int argc, char **argv)
 	g_test_add_func("/testsms/Test Deliver Encode", test_deliver_encode);
 	g_test_add_func("/testsms/Test Simple Submit", test_simple_submit);
 	g_test_add_func("/testsms/Test Submit Encode", test_submit_encode);
-	g_test_add_func("/testsms/Test UDH Iterator", test_udh_iter);
+
+	g_test_add_data_func("/testsms/Test EMS UDH 1",
+			&ems_udh_test_1, test_ems_udh);
+	g_test_add_data_func("/testsms/Test EMS UDH 2",
+			&ems_udh_test_2, test_ems_udh);
+
 	g_test_add_func("/testsms/Test Assembly", test_assembly);
 	g_test_add_func("/testsms/Test Prepare 7Bit", test_prepare_7bit);
-	g_test_add_func("/testsms/Test Prepare Concat", test_prepare_concat);
+
+	g_test_add_data_func("/testsms/Test Prepare Concat",
+			&shakespeare_test, test_prepare_concat);
+
+	memset(long_string, 'a', 152*30);
+	memset(long_string + 152*30, 'b', 152);
+	memset(long_string + 152*31, 'c', 152);
+	memset(long_string + 152*32, 'd', 152);
+	long_string[152*33] = '\0';
+
+	long_string_test.str = long_string;
+	long_string_test.segments = 33;
+
+	g_test_add_data_func("/testsms/Test Prepare Concat 30+ segments",
+			&long_string_test, test_prepare_concat);
+
 	g_test_add_func("/testsms/Test Prepare Limits", test_prepare_limits);
 
 	g_test_add_func("/testsms/Test CBS Encode / Decode",
@@ -1182,5 +1383,7 @@ int main(int argc, char **argv)
 
 	g_test_add_func("/testsms/Range minimizer", test_range_minimizer);
 
+	g_test_add_func("/testsms/Status Report Assembly", test_sr_assembly);
+
 	return g_test_run();
 }
